from sqlalchemy import Column, Integer, String, ForeignKey, Boolean, DateTime, LargeBinary  # type: ignore

from sqlalchemy.orm import relationship, backref  # type: ignore

from sqlalchemy.ext.declarative import declarative_base  # type: ignore


Base = declarative_base()


# === User Data ===

class Document(Base):
    __tablename__ = "documents"
    id = Column(Integer, primary_key=True)
    filepath = Column(String, nullable=False, unique=True)
    title = Column(String)
    author = Column(String)
    cover_data = Column(LargeBinary)
    cover_width = Column(Integer)
    cover_height = Column(Integer)
    cover_stride = Column(Integer)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)
    notes = relationship("Note", backref=backref("note"))


class Note(Base):
    __tablename__ = "notes"
    id = Column(Integer, primary_key=True)
    document_id = Column(Integer, ForeignKey("documents.id"))
    doc_page_number = Column(Integer)
    front = Column(String)
    back = Column(String)
    anki_note_id = Column(Integer)
    anki_synced_at = Column(DateTime)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)


USERDATA_CREATE_SCHEMA_SQL = """
CREATE TABLE `documents` (
  `id`           INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  `filepath`     TEXT NOT NULL UNIQUE,
  `title`        VARCHAR,
  `author`       VARCHAR,
  `cover_data`   BLOB,
  `cover_width`  INTEGER,
  `cover_height` INTEGER,
  `cover_stride` INTEGER,
  `created_at`      TEXT DEFAULT CURRENT_TIMESTAMP,
  `updated_at`      TEXT
);

CREATE TABLE `notes` (
  `id`              INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  `document_id`     INTEGER REFERENCES `documents` (`id`) ON DELETE NO ACTION ON UPDATE CASCADE,
  `doc_page_number` INTEGER,
  `front`           TEXT,
  `back`            TEXT,
  `anki_note_id`    INTEGER,
  `anki_synced_at`  TEXT,
  `created_at`      TEXT DEFAULT CURRENT_TIMESTAMP,
  `updated_at`      TEXT
);
"""


# === App Data ===

class Author(Base):
    __tablename__ = "authors"
    id = Column(Integer, primary_key=True)
    uid = Column(String)
    blurb = Column(String)
    long_name = Column(String)
    short_name = Column(String)
    root_texts = relationship("RootText", backref=backref("author"))


class RootText(Base):
    __tablename__ = "root_texts"
    id = Column(Integer, primary_key=True)
    author_id = Column(Integer, ForeignKey("authors.id"))
    uid = Column(String)
    acronym = Column(String)
    volpage = Column(String)
    title = Column(String)
    content_language = Column(String)
    content_plain = Column(String)
    content_html = Column(String)


class Dictionary(Base):
    __tablename__ = "dictionaries"
    id = Column(Integer, primary_key=True)
    label = Column(String)
    title = Column(String)


class DictWord(Base):
    __tablename__ = "dict_words"
    id = Column(Integer, primary_key=True)
    dictionary_id = Column(Integer, ForeignKey("dictionaries.id"))
    word = Column(String)
    word_nom_sg = Column(String)
    inflections = Column(String)
    phonetic = Column(String)
    transliteration = Column(String)
    url_id = Column(String)
    meanings = relationship("Meaning", backref=backref("dict_word"))


class Meaning(Base):
    __tablename__ = "meanings"
    id = Column(Integer, primary_key=True)
    dict_word_id = Column(Integer, ForeignKey("dict_words.id"))
    meaning_order = Column(Integer)
    definition_md = Column(String)
    summary = Column(String)
    synonyms = Column(String)
    antonyms = Column(String)
    homonyms = Column(String)
    also_written_as = Column(String)
    see_also = Column(String)
    comment = Column(String)
    is_root = Column(Boolean)
    root_language = Column(String)
    root_groups = Column(String)
    root_sign = Column(String)
    root_numbered_group = Column(String)





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'simsapa/assets/ui/sutta_search_window.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_SuttaSearchWindow(object):
    def setupUi(self, SuttaSearchWindow):
        SuttaSearchWindow.setObjectName("SuttaSearchWindow")
        SuttaSearchWindow.resize(856, 581)
        SuttaSearchWindow.setBaseSize(QtCore.QSize(800, 600))
        self.central_widget = QtWidgets.QWidget(SuttaSearchWindow)
        self.central_widget.setObjectName("central_widget")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.central_widget)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setObjectName("main_layout")
        self.searchbar_layout = QtWidgets.QHBoxLayout()
        self.searchbar_layout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.searchbar_layout.setObjectName("searchbar_layout")
        self.search_input = QtWidgets.QLineEdit(self.central_widget)
        self.search_input.setMinimumSize(QtCore.QSize(0, 35))
        self.search_input.setClearButtonEnabled(True)
        self.search_input.setObjectName("search_input")
        self.searchbar_layout.addWidget(self.search_input)
        self.search_button = QtWidgets.QPushButton(self.central_widget)
        self.search_button.setMinimumSize(QtCore.QSize(0, 40))
        self.search_button.setObjectName("search_button")
        self.searchbar_layout.addWidget(self.search_button)
        self.main_layout.addLayout(self.searchbar_layout)
        self.splitter = QtWidgets.QSplitter(self.central_widget)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.verticalLayoutWidget_2 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_2.setObjectName("verticalLayoutWidget_2")
        self.content_layout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_2)
        self.content_layout.setContentsMargins(0, 0, 0, 0)
        self.content_layout.setObjectName("content_layout")
        self.content_title = QtWidgets.QLabel(self.verticalLayoutWidget_2)
        self.content_title.setObjectName("content_title")
        self.content_layout.addWidget(self.content_title)
        self.content_html = QtWidgets.QTextBrowser(self.verticalLayoutWidget_2)
        self.content_html.setObjectName("content_html")
        self.content_layout.addWidget(self.content_html)
        self.verticalLayoutWidget_3 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_3.setObjectName("verticalLayoutWidget_3")
        self.results_layout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_3)
        self.results_layout.setContentsMargins(0, 0, 0, 0)
        self.results_layout.setObjectName("results_layout")
        self.label_2 = QtWidgets.QLabel(self.verticalLayoutWidget_3)
        self.label_2.setObjectName("label_2")
        self.results_layout.addWidget(self.label_2)
        self.results_list = QtWidgets.QListWidget(self.verticalLayoutWidget_3)
        self.results_list.setObjectName("results_list")
        self.results_layout.addWidget(self.results_list)
        self.label_3 = QtWidgets.QLabel(self.verticalLayoutWidget_3)
        self.label_3.setObjectName("label_3")
        self.results_layout.addWidget(self.label_3)
        self.history_list = QtWidgets.QListWidget(self.verticalLayoutWidget_3)
        self.history_list.setObjectName("history_list")
        self.results_layout.addWidget(self.history_list)
        self.main_layout.addWidget(self.splitter)
        self.horizontalLayout_2.addLayout(self.main_layout)
        SuttaSearchWindow.setCentralWidget(self.central_widget)
        self.menubar = QtWidgets.QMenuBar(SuttaSearchWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 856, 20))
        self.menubar.setObjectName("menubar")
        self.menu_File = QtWidgets.QMenu(self.menubar)
        self.menu_File.setObjectName("menu_File")
        self.menu_Edit = QtWidgets.QMenu(self.menubar)
        self.menu_Edit.setObjectName("menu_Edit")
        self.menu_Windows = QtWidgets.QMenu(self.menubar)
        self.menu_Windows.setObjectName("menu_Windows")
        self.menu_Help = QtWidgets.QMenu(self.menubar)
        self.menu_Help.setObjectName("menu_Help")
        SuttaSearchWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(SuttaSearchWindow)
        self.statusbar.setObjectName("statusbar")
        SuttaSearchWindow.setStatusBar(self.statusbar)
        self.toolBar = QtWidgets.QToolBar(SuttaSearchWindow)
        self.toolBar.setObjectName("toolBar")
        SuttaSearchWindow.addToolBar(QtCore.Qt.LeftToolBarArea, self.toolBar)
        self.action_Copy = QtWidgets.QAction(SuttaSearchWindow)
        self.action_Copy.setObjectName("action_Copy")
        self.action_Paste = QtWidgets.QAction(SuttaSearchWindow)
        self.action_Paste.setObjectName("action_Paste")
        self.action_Quit = QtWidgets.QAction(SuttaSearchWindow)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/close"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Quit.setIcon(icon)
        self.action_Quit.setObjectName("action_Quit")
        self.action_Sutta_Search = QtWidgets.QAction(SuttaSearchWindow)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/book"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Sutta_Search.setIcon(icon1)
        self.action_Sutta_Search.setObjectName("action_Sutta_Search")
        self.action_Dictionary_Search = QtWidgets.QAction(SuttaSearchWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/dictionary"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Dictionary_Search.setIcon(icon2)
        self.action_Dictionary_Search.setObjectName("action_Dictionary_Search")
        self.action_About = QtWidgets.QAction(SuttaSearchWindow)
        self.action_About.setObjectName("action_About")
        self.action_Website = QtWidgets.QAction(SuttaSearchWindow)
        self.action_Website.setObjectName("action_Website")
        self.action_Close_Window = QtWidgets.QAction(SuttaSearchWindow)
        self.action_Close_Window.setObjectName("action_Close_Window")
        self.action_Open = QtWidgets.QAction(SuttaSearchWindow)
        self.action_Open.setObjectName("action_Open")
        self.action_Document_Reader = QtWidgets.QAction(SuttaSearchWindow)
        self.action_Document_Reader.setObjectName("action_Document_Reader")
        self.action_Library = QtWidgets.QAction(SuttaSearchWindow)
        self.action_Library.setObjectName("action_Library")
        self.action_Notes = QtWidgets.QAction(SuttaSearchWindow)
        self.action_Notes.setObjectName("action_Notes")
        self.menu_File.addAction(self.action_Open)
        self.menu_File.addAction(self.action_Close_Window)
        self.menu_File.addSeparator()
        self.menu_File.addAction(self.action_Quit)
        self.menu_Edit.addAction(self.action_Copy)
        self.menu_Edit.addAction(self.action_Paste)
        self.menu_Windows.addAction(self.action_Sutta_Search)
        self.menu_Windows.addAction(self.action_Dictionary_Search)
        self.menu_Windows.addAction(self.action_Document_Reader)
        self.menu_Windows.addAction(self.action_Library)
        self.menu_Windows.addAction(self.action_Notes)
        self.menu_Help.addAction(self.action_Website)
        self.menu_Help.addAction(self.action_About)
        self.menubar.addAction(self.menu_File.menuAction())
        self.menubar.addAction(self.menu_Edit.menuAction())
        self.menubar.addAction(self.menu_Windows.menuAction())
        self.menubar.addAction(self.menu_Help.menuAction())
        self.toolBar.addAction(self.action_Sutta_Search)
        self.toolBar.addAction(self.action_Dictionary_Search)

        self.retranslateUi(SuttaSearchWindow)
        QtCore.QMetaObject.connectSlotsByName(SuttaSearchWindow)

    def retranslateUi(self, SuttaSearchWindow):
        _translate = QtCore.QCoreApplication.translate
        SuttaSearchWindow.setWindowTitle(_translate("SuttaSearchWindow", "Sutta Search - Simsapa"))
        self.search_button.setText(_translate("SuttaSearchWindow", "Search"))
        self.content_title.setText(_translate("SuttaSearchWindow", "Title"))
        self.label_2.setText(_translate("SuttaSearchWindow", "Results"))
        self.label_3.setText(_translate("SuttaSearchWindow", "History"))
        self.menu_File.setTitle(_translate("SuttaSearchWindow", "&File"))
        self.menu_Edit.setTitle(_translate("SuttaSearchWindow", "&Edit"))
        self.menu_Windows.setTitle(_translate("SuttaSearchWindow", "&Windows"))
        self.menu_Help.setTitle(_translate("SuttaSearchWindow", "&Help"))
        self.toolBar.setWindowTitle(_translate("SuttaSearchWindow", "toolBar"))
        self.action_Copy.setText(_translate("SuttaSearchWindow", "&Copy"))
        self.action_Paste.setText(_translate("SuttaSearchWindow", "&Paste"))
        self.action_Quit.setText(_translate("SuttaSearchWindow", "&Quit"))
        self.action_Quit.setShortcut(_translate("SuttaSearchWindow", "Ctrl+Q"))
        self.action_Sutta_Search.setText(_translate("SuttaSearchWindow", "&Sutta Search"))
        self.action_Sutta_Search.setShortcut(_translate("SuttaSearchWindow", "F5"))
        self.action_Dictionary_Search.setText(_translate("SuttaSearchWindow", "&Dictionary Search"))
        self.action_Dictionary_Search.setShortcut(_translate("SuttaSearchWindow", "F6"))
        self.action_About.setText(_translate("SuttaSearchWindow", "&About"))
        self.action_Website.setText(_translate("SuttaSearchWindow", "&Website"))
        self.action_Close_Window.setText(_translate("SuttaSearchWindow", "&Close Window"))
        self.action_Open.setText(_translate("SuttaSearchWindow", "&Open..."))
        self.action_Open.setShortcut(_translate("SuttaSearchWindow", "Ctrl+O"))
        self.action_Document_Reader.setText(_translate("SuttaSearchWindow", "D&ocument Reader"))
        self.action_Document_Reader.setToolTip(_translate("SuttaSearchWindow", "Document Reader"))
        self.action_Document_Reader.setShortcut(_translate("SuttaSearchWindow", "F7"))
        self.action_Library.setText(_translate("SuttaSearchWindow", "&Library"))
        self.action_Library.setShortcut(_translate("SuttaSearchWindow", "F8"))
        self.action_Notes.setText(_translate("SuttaSearchWindow", "&Notes"))
        self.action_Notes.setToolTip(_translate("SuttaSearchWindow", "Notes"))
        self.action_Notes.setShortcut(_translate("SuttaSearchWindow", "F9"))
from simsapa.assets import icons_rc



import os.path
import epub_meta
from PyPDF2 import PdfFileReader
from typing import Optional

try:
    import fitz  # type: ignore
    has_fitz = True
except ImportError:
    has_fitz = False

import logging as _logging

logger = _logging.getLogger(__name__)


class PageImage():
    def __init__(self, file_doc, page_idx: int, zoom: Optional[float] = None):
        if not has_fitz:
            return None

        if file_doc is None or page_idx > file_doc.number_of_pages() - 1:
            return None

        self.zoom = zoom or file_doc._zoom
        self.pixmap = file_doc.file_doc[page_idx] \
                          .get_pixmap(
                              matrix=fitz.Matrix(self.zoom, self.zoom),
                              alpha=False
                          )

        self.image_bytes = self.pixmap.tobytes("ppm")

        self.width = self.pixmap.width
        self.height = self.pixmap.height
        self.stride = self.pixmap.stride


class FileDoc():
    def __init__(self, filepath):

        if not os.path.exists(filepath):
            logger.error(f"File not found: {filepath}")

        self.filepath = filepath
        name, ext = os.path.splitext(self.filepath)
        self.file_name = name
        self.file_ext = ext.lower()
        self.title = ""
        self.author = ""

        self._current_idx: int = 0
        self._zoom = 1.5

        if has_fitz:
            self.file_doc = fitz.open(filepath)
            self._matrix = fitz.Matrix(self._zoom, self._zoom)
        elif self.file_ext == ".pdf":
            self.file_doc = PdfFileReader(open(self.filepath, "rb"))
        else:
            self.file_doc = None

        self._parse_metadata()

    def number_of_pages(self) -> int:
        if has_fitz and self.file_doc:
            return len(self.file_doc)
        else:
            return 0

    def set_page_number(self, page: int):
        if page > 0:
            self._current_idx = page - 1

    def current_page_number(self) -> int:
        return self._current_idx + 1

    def current_page_idx(self) -> int:
        return self._current_idx

    def current_page_image(self) -> Optional[PageImage]:
        return self.page_image(self._current_idx)

    def page_image(self, page_idx: int, zoom: Optional[float] = None) -> Optional[PageImage]:
        if not has_fitz:
            return None

        return PageImage(self, page_idx, zoom)

    def _parse_metadata(self):
        if not self.file_doc:
            return

        if has_fitz:
            self.title = self.file_doc.metadata['title']
            self.author = self.file_doc.metadata['author']

        elif self.file_ext == ".pdf":
            info = self.file_doc.getDocumentInfo()
            self.title = info.title
            self.author = info.author

        elif self.file_ext == ".epub":
            info = epub_meta.get_epub_metadata(self.filepath, read_cover_image=False, read_toc=False)
            self.title = info['title']
            self.author = info['authors'][0]





import logging as _logging
import os.path
from functools import partial
from typing import List, Optional
from sqlalchemy.sql import func  # type: ignore

from PyQt5.QtCore import QAbstractListModel, Qt  # type: ignore
from PyQt5.QtGui import QIcon, QImage, QPixmap  # type: ignore
from PyQt5.QtWidgets import (QFileDialog, QLabel, QMainWindow,  # type: ignore
                             QMessageBox)
from simsapa.assets import icons_rc

from ..app.file_doc import FileDoc  # type: ignore
from ..app.db_models import Document  # type: ignore
from ..app.types import AppData  # type: ignore
from ..assets.ui.library_browser_window_ui import Ui_LibraryBrowserWindow  # type: ignore

logger = _logging.getLogger(__name__)


class DocumentListModel(QAbstractListModel):
    def __init__(self, *args, docs=None, **kwargs):
        super(DocumentListModel, self).__init__(*args, **kwargs)
        self.docs = docs or []

    def data(self, index, role):
        if role == Qt.DisplayRole:
            filepath = self.docs[index.row()].filepath
            name = os.path.basename(filepath)
            return name

        if role == Qt.DecorationRole:
            filepath = self.docs[index.row()].filepath
            if not os.path.exists(filepath):
                return QIcon(":close")

    def rowCount(self, index):
        return len(self.docs)


class LibraryBrowserWindow(QMainWindow, Ui_LibraryBrowserWindow):
    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self._app_data: AppData = app_data

        self.model = DocumentListModel(docs=self._get_all_documents())

        self.documents_list.setModel(self.model)
        self.sel_model = self.documents_list.selectionModel()

        self._ui_setup()
        self._connect_signals()

        self.statusbar.showMessage("Ready", 3000)

    def _ui_setup(self):
        self.status_msg = QLabel("")
        self.statusbar.addPermanentWidget(self.status_msg)

        self.search_input.setFocus()
        self._show_document_clear()

    def _get_all_documents(self) -> List[Document]:
        return self._app_data.user_db_session.query(Document).all()

    def _handle_query(self):
        pass

    def get_selected_document(self) -> Optional[Document]:
        a = self.documents_list.selectedIndexes()
        if not a:
            return None

        item = a[0]
        return self.model.docs[item.row()]

    def remove_selected_document(self):
        a = self.documents_list.selectedIndexes()
        if not a:
            return None

        # Remove from model
        item = a[0]
        doc_id = self.model.docs[item.row()].id

        del self.model.docs[item.row()]
        self.model.layoutChanged.emit()
        self.documents_list.clearSelection()
        self._show_document_clear()

        # Remove from database

        db_item = self._app_data.user_db_session \
                                .query(Document) \
                                .filter(Document.id == doc_id) \
                                .first()
        self._app_data.user_db_session.delete(db_item)
        self._app_data.user_db_session.commit()

    def _handle_document_select(self):
        doc = self.get_selected_document()
        if doc:
            self._show_document(doc)

    def _show_document_clear(self):
        self.status_msg.clear()
        self.doc_title.clear()
        self.doc_author.clear()
        self.doc_cover.clear()

    def _show_document(self, doc: Document):
        self.status_msg.setText(doc.filepath)
        self.doc_title.setText(doc.title)
        self.doc_author.setText(doc.author)

        if doc.cover_data:
            img = QImage(
                doc.cover_data,
                doc.cover_width,
                doc.cover_height,
                doc.cover_stride,
                QImage.Format.Format_RGB888
            )
            self.doc_cover.setPixmap(QPixmap.fromImage(img))

    def _documents_search_query(self, query: str):
        results = self._app_data.user_db_session \
                               .query(Document) \
                               .filter(Document.filepath.like(f"%{query}%")) \
                               .all()
        return results

    def add_document(self, filepath):
        if not os.path.exists(filepath):
            logger.error(f"File doesn't exist: {filepath}")
            return

        file_doc = FileDoc(filepath)

        title = file_doc.title or "Unknown"

        author = file_doc.author or "Unknown"

        page_image = file_doc.page_image(0, 0.7)

        # Insert or update database record

        item = self._app_data.user_db_session \
                             .query(Document) \
                             .filter(Document.filepath == filepath) \
                             .first()
        if item is None:
            logger.info(f"Add new: {filepath}")

            if page_image:
                db_doc = Document(
                    filepath=filepath,
                    title=title,
                    author=author,
                    cover_data=page_image.image_bytes,
                    cover_width=page_image.width,
                    cover_height=page_image.height,
                    cover_stride=page_image.stride,
                    created_at=func.now(),
                )
            else:
                db_doc = Document(
                    filepath=filepath,
                    title=title,
                    author=author,
                    created_at=func.now(),
                )

            try:
                self._app_data.user_db_session.add(db_doc)
                self._app_data.user_db_session.commit()

                # Add to model
                self.model.docs.append(db_doc)

            except Exception as e:
                logger.error(e)

        else:
            logger.info(f"Update: {item.filepath}")

            if page_image:
                values = {
                    'filepath': filepath,
                    'title': title,
                    'author': author,
                    'cover_data': page_image.image_bytes,
                    'cover_width': page_image.width,
                    'cover_height': page_image.height,
                    'cover_stride': page_image.stride,
                    'updated_at': func.now(),
                }
            else:
                values = {
                    'filepath': filepath,
                    'title': title,
                    'author': author,
                    'updated_at': func.now(),
                }

            try:
                self._app_data.user_db_session \
                              .query(Document) \
                              .filter(Document.id == item.id) \
                              .update(values)

                self._app_data.user_db_session.commit()

            except Exception as e:
                logger.error(e)

        self.model.layoutChanged.emit()

    def _add_document_dialog(self):
        file_paths, _ = QFileDialog.getOpenFileNames(
            None,
            "Add Documents to the Library...",
            "",
            "PDF or Epub Files (*.pdf *.epub)")

        for p in file_paths:
            self.add_document(p)

    def _remove_document_dialog(self):
        doc = self.get_selected_document()
        if not doc:
            return

        reply = QMessageBox.question(self,
                                     'Remove Document...',
                                     'Remove this item from the Library? (Files are NOT deleted.)',
                                     QMessageBox.Yes | QMessageBox.No,
                                     QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.remove_selected_document()

    def _connect_signals(self):
        self.action_Close_Window \
            .triggered.connect(partial(self.close))

        self.action_Add \
            .triggered.connect(partial(self._add_document_dialog))

        self.action_Remove \
            .triggered.connect(partial(self._remove_document_dialog))

        self.search_button.clicked.connect(partial(self._handle_query))
        self.search_input.textChanged.connect(partial(self._handle_query))

        self.sel_model.selectionChanged.connect(partial(self._handle_document_select))



from functools import partial
from typing import List, Optional
from sqlalchemy.sql import func  # type: ignore
import logging as _logging

from PyQt5.QtCore import QAbstractListModel, Qt
from PyQt5.QtGui import QImage, QPixmap
from PyQt5.QtWidgets import (QLabel, QMainWindow, QFileDialog, QInputDialog,
                             QMessageBox)  # type: ignore

from ..app.file_doc import FileDoc  # type: ignore
from ..app.db_models import Document, Note  # type: ignore

from ..app.types import AppData  # type: ignore
from ..assets.ui.document_reader_window_ui import Ui_DocumentReaderWindow  # type: ignore

logger = _logging.getLogger(__name__)


class NoteListModel(QAbstractListModel):
    def __init__(self, *args, notes=None, **kwargs):
        super(NoteListModel, self).__init__(*args, **kwargs)
        self.notes = notes or []

    def data(self, index, role):
        if role == Qt.DisplayRole:
            text = self.notes[index.row()].front + " " + self.notes[index.row()].back
            return text

    def rowCount(self, index):
        if self.notes:
            return len(self.notes)
        else:
            return 0


class DocumentReaderWindow(QMainWindow, Ui_DocumentReaderWindow):
    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self._app_data: AppData = app_data

        self.model = NoteListModel()
        self.notes_list.setModel(self.model)
        self.sel_model = self.notes_list.selectionModel()

        self._ui_setup()

        self.statusbar.showMessage("Ready", 3000)

    def _ui_setup(self):
        self.status_msg = QLabel("")
        self.statusbar.addPermanentWidget(self.status_msg)
        self._show_note_clear()

        self.file_doc = None
        self.db_doc = None

    def open_file_dialog(self):
        file_path, _ = QFileDialog.getOpenFileName(
            None,
            "Open File...",
            "",
            "PDF or Epub Files (*.pdf *.epub)")

        if len(file_path) != 0:
            self.open_doc(file_path)

    def open_doc(self, path):
        self.db_doc = self._app_data.user_db_session \
                                .query(Document) \
                                .filter(Document.filepath == path) \
                                .first()

        notes = self._get_notes_for_this_page()
        if notes:
            self.model.notes = notes
        else:
            self.model.notes = []

        self.model.layoutChanged.emit()

        self.file_doc = FileDoc(path)
        self.doc_go_to_page(1)

    def doc_show_current(self):
        self.doc_go_to_page(self.file_doc.current_page_number())

    def doc_go_to_page(self, page: int):
        if not self.file_doc:
            return

        logger.info(f"doc_go_to_page({page})")

        page_count = self.file_doc.number_of_pages()

        if self.file_doc is None or page < 1 or page > page_count:
            return

        self.page_current_of_total.setText(f"{page} of {page_count}")

        if self.file_doc:
            self.file_doc.set_page_number(page)

            page_img = self.file_doc.current_page_image()
            if page_img:
                img = QImage(page_img.image_bytes, page_img.width, page_img.height, page_img.stride, QImage.Format.Format_RGB888)
                self.content_page.setPixmap(QPixmap.fromImage(img))

        if self.db_doc:
            self.notes_list.clearSelection()
            self._show_note_clear()
            self.model.notes = self._get_notes_for_this_page()
            self.model.layoutChanged.emit()

    def _previous_page(self):
        page_nr = self.file_doc.current_page_number() - 1
        if self.file_doc and page_nr > 0:
            self.file_doc.set_page_number(page_nr)
            self._upd_current_page_input(page_nr)

    def _next_page(self):
        page_nr = self.file_doc.current_page_number() + 1
        if self.file_doc and page_nr <= self.file_doc.number_of_pages():
            self.file_doc.set_page_number(page_nr)
            self._upd_current_page_input(page_nr)

    def _beginning(self):
        if self.file_doc and self.file_doc.current_page_number() != 1:
            self.file_doc.set_page_number(1)
            self._upd_current_page_input(1)

    def _end(self):
        if self.file_doc and self.file_doc.current_page_number() != self.file_doc.number_of_pages():
            self.file_doc.set_page_number(self.file_doc.number_of_pages())
            self._upd_current_page_input(self.file_doc.number_of_pages())

    def _upd_current_page_input(self, n):
        self.current_page_input.setValue(n)
        self.current_page_input.clearFocus()

    def _go_to_page_dialog(self):
        n, ok = QInputDialog.getInt(self, "Go to Page...", "Page:", 1, 1, self.file_doc.number_of_pages(), 1)
        if ok:
            self._upd_current_page_input(n)

    def _go_to_page_input(self):
        n = self.current_page_input.value()
        self.doc_go_to_page(n)

    def _get_notes_for_this_page(self) -> List[Note]:
        if self.db_doc is None or self.file_doc is None:
            return

        notes = self._app_data.user_db_session \
                             .query(Note) \
                             .filter(
                                 Note.document_id == self.db_doc.id,
                                 Note.doc_page_number == self.file_doc.current_page_number()) \
                             .all()

        return notes

    def get_selected_note(self) -> Optional[Note]:
        a = self.notes_list.selectedIndexes()
        if not a:
            return None

        item = a[0]
        return self.model.notes[item.row()]

    def remove_selected_note(self):
        a = self.notes_list.selectedIndexes()
        if not a:
            return None

        # Remove from model
        item = a[0]
        note_id = self.model.notes[item.row()].id

        del self.model.notes[item.row()]
        self.model.layoutChanged.emit()
        self.notes_list.clearSelection()
        self._show_note_clear()

        # Remove from database

        db_item = self._app_data.user_db_session \
                                .query(Note) \
                                .filter(Note.id == note_id) \
                                .first()
        self._app_data.user_db_session.delete(db_item)
        self._app_data.user_db_session.commit()

    def _handle_note_select(self):
        note = self.get_selected_note()
        if note:
            self._show_note(note)

    def _show_note_clear(self):
        self.front_input.clear()
        self.back_input.clear()

    def _show_note(self, note: Note):
        self.front_input.setPlainText(note.front)
        self.back_input.setPlainText(note.back)

    def add_note(self):

        front = self.front_input.toPlainText()
        back = self.back_input.toPlainText()

        if len(front) == 0 and len(back) == 0:
            logger.info("Empty content, cancel adding.")
            return

        # Insert database record

        logger.info(f"Adding new note")

        if self.db_doc is None:
            db_note = Note(
                front=front,
                back=back
            )
        else:
            db_note = Note(
                front=front,
                back=back,
                document_id=self.db_doc.id,
                doc_page_number=self.file_doc.current_page_number(),
            )

        try:
            self._app_data.user_db_session.add(db_note)
            self._app_data.user_db_session.commit()

            # Add to model
            if self.model.notes:
                self.model.notes.append(db_note)
            else:
                self.model.notes = [db_note]

        except Exception as e:
            logger.error(e)

        self.model.layoutChanged.emit()

    def update_selected_note_front(self):
        note = self.get_selected_note()
        if note is None:
            return

        note.front = self.front_input.toPlainText()

        self._app_data.user_db_session.commit()
        self.model.layoutChanged.emit()

    def update_selected_note_back(self):
        note = self.get_selected_note()
        if note is None:
            return

        note.back = self.back_input.toPlainText()

        self._app_data.user_db_session.commit()
        self.model.layoutChanged.emit()

    def remove_note_dialog(self):
        note = self.get_selected_note()
        if not note:
            return

        reply = QMessageBox.question(self,
                                     'Remove Note...',
                                     'Remove this item?',
                                     QMessageBox.Yes | QMessageBox.No,
                                     QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.remove_selected_note()

class DocumentReaderCtrl:
    def __init__(self, view):
        self._view = view
        self._connect_signals()

    def _connect_signals(self):
        self._view.action_Close_Window \
            .triggered.connect(partial(self._view.close))

        self._view.action_Previous_Page \
            .triggered.connect(partial(self._view._previous_page))
        self._view.action_Next_Page \
            .triggered.connect(partial(self._view._next_page))

        self._view.action_Beginning \
            .triggered.connect(partial(self._view._beginning))
        self._view.action_End \
            .triggered.connect(partial(self._view._end))

        self._view.action_Go_to_Page \
            .triggered.connect(partial(self._view._go_to_page_dialog))

        self._view.current_page_input \
            .valueChanged.connect(partial(self._view._go_to_page_input))

        self._view.sel_model.selectionChanged.connect(partial(self._view._handle_note_select))

        self._view.add_note_button \
                  .clicked.connect(partial(self._view.add_note))

        self._view.remove_note_button \
                  .clicked.connect(partial(self._view.remove_note_dialog))

        self._view.front_input \
                  .textChanged.connect(partial(self._view.update_selected_note_front))

        self._view.back_input \
                  .textChanged.connect(partial(self._view.update_selected_note_back))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'simsapa/assets/ui/library_browser_window.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_LibraryBrowserWindow(object):
    def setupUi(self, LibraryBrowserWindow):
        LibraryBrowserWindow.setObjectName("LibraryBrowserWindow")
        LibraryBrowserWindow.resize(856, 581)
        LibraryBrowserWindow.setBaseSize(QtCore.QSize(800, 600))
        self.central_widget = QtWidgets.QWidget(LibraryBrowserWindow)
        self.central_widget.setObjectName("central_widget")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.central_widget)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setObjectName("main_layout")
        self.searchbar_layout = QtWidgets.QHBoxLayout()
        self.searchbar_layout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.searchbar_layout.setObjectName("searchbar_layout")
        self.search_input = QtWidgets.QLineEdit(self.central_widget)
        self.search_input.setMinimumSize(QtCore.QSize(0, 35))
        self.search_input.setClearButtonEnabled(True)
        self.search_input.setObjectName("search_input")
        self.searchbar_layout.addWidget(self.search_input)
        self.search_button = QtWidgets.QPushButton(self.central_widget)
        self.search_button.setMinimumSize(QtCore.QSize(0, 40))
        self.search_button.setObjectName("search_button")
        self.searchbar_layout.addWidget(self.search_button)
        self.main_layout.addLayout(self.searchbar_layout)
        self.splitter = QtWidgets.QSplitter(self.central_widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.splitter.sizePolicy().hasHeightForWidth())
        self.splitter.setSizePolicy(sizePolicy)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.verticalLayoutWidget_2 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_2.setObjectName("verticalLayoutWidget_2")
        self.documents_layout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_2)
        self.documents_layout.setContentsMargins(0, 0, 0, 0)
        self.documents_layout.setObjectName("documents_layout")
        self.documents_list = QtWidgets.QListView(self.verticalLayoutWidget_2)
        self.documents_list.setObjectName("documents_list")
        self.documents_layout.addWidget(self.documents_list)
        self.verticalLayoutWidget_3 = QtWidgets.QWidget(self.splitter)
        self.verticalLayoutWidget_3.setObjectName("verticalLayoutWidget_3")
        self.info_layout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget_3)
        self.info_layout.setContentsMargins(0, 0, 0, 0)
        self.info_layout.setObjectName("info_layout")
        self.doc_cover = QtWidgets.QLabel(self.verticalLayoutWidget_3)
        self.doc_cover.setMinimumSize(QtCore.QSize(100, 0))
        self.doc_cover.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.doc_cover.setObjectName("doc_cover")
        self.info_layout.addWidget(self.doc_cover)
        self.doc_title = QtWidgets.QLabel(self.verticalLayoutWidget_3)
        self.doc_title.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.doc_title.setObjectName("doc_title")
        self.info_layout.addWidget(self.doc_title)
        self.doc_author = QtWidgets.QLabel(self.verticalLayoutWidget_3)
        self.doc_author.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.doc_author.setObjectName("doc_author")
        self.info_layout.addWidget(self.doc_author)
        self.label = QtWidgets.QLabel(self.verticalLayoutWidget_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setText("")
        self.label.setObjectName("label")
        self.info_layout.addWidget(self.label)
        self.main_layout.addWidget(self.splitter)
        self.horizontalLayout_2.addLayout(self.main_layout)
        LibraryBrowserWindow.setCentralWidget(self.central_widget)
        self.menubar = QtWidgets.QMenuBar(LibraryBrowserWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 856, 20))
        self.menubar.setObjectName("menubar")
        self.menu_File = QtWidgets.QMenu(self.menubar)
        self.menu_File.setObjectName("menu_File")
        self.menu_Edit = QtWidgets.QMenu(self.menubar)
        self.menu_Edit.setObjectName("menu_Edit")
        self.menu_Windows = QtWidgets.QMenu(self.menubar)
        self.menu_Windows.setObjectName("menu_Windows")
        self.menu_Help = QtWidgets.QMenu(self.menubar)
        self.menu_Help.setObjectName("menu_Help")
        self.menu_Library = QtWidgets.QMenu(self.menubar)
        self.menu_Library.setObjectName("menu_Library")
        LibraryBrowserWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(LibraryBrowserWindow)
        self.statusbar.setObjectName("statusbar")
        LibraryBrowserWindow.setStatusBar(self.statusbar)
        self.toolBar = QtWidgets.QToolBar(LibraryBrowserWindow)
        self.toolBar.setObjectName("toolBar")
        LibraryBrowserWindow.addToolBar(QtCore.Qt.LeftToolBarArea, self.toolBar)
        self.toolBar_2 = QtWidgets.QToolBar(LibraryBrowserWindow)
        self.toolBar_2.setObjectName("toolBar_2")
        LibraryBrowserWindow.addToolBar(QtCore.Qt.TopToolBarArea, self.toolBar_2)
        self.action_Copy = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Copy.setObjectName("action_Copy")
        self.action_Paste = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Paste.setObjectName("action_Paste")
        self.action_Quit = QtWidgets.QAction(LibraryBrowserWindow)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/close"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Quit.setIcon(icon)
        self.action_Quit.setObjectName("action_Quit")
        self.action_Sutta_Search = QtWidgets.QAction(LibraryBrowserWindow)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/book"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Sutta_Search.setIcon(icon1)
        self.action_Sutta_Search.setObjectName("action_Sutta_Search")
        self.action_Dictionary_Search = QtWidgets.QAction(LibraryBrowserWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/dictionary"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.action_Dictionary_Search.setIcon(icon2)
        self.action_Dictionary_Search.setObjectName("action_Dictionary_Search")
        self.action_About = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_About.setObjectName("action_About")
        self.action_Website = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Website.setObjectName("action_Website")
        self.action_Close_Window = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Close_Window.setObjectName("action_Close_Window")
        self.action_Open = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Open.setObjectName("action_Open")
        self.action_Document_Reader = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Document_Reader.setObjectName("action_Document_Reader")
        self.action_Add = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Add.setObjectName("action_Add")
        self.action_Remove = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Remove.setObjectName("action_Remove")
        self.action_Open_Selected = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Open_Selected.setObjectName("action_Open_Selected")
        self.action_Library = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Library.setObjectName("action_Library")
        self.action_Notes = QtWidgets.QAction(LibraryBrowserWindow)
        self.action_Notes.setObjectName("action_Notes")
        self.menu_File.addAction(self.action_Open)
        self.menu_File.addAction(self.action_Close_Window)
        self.menu_File.addSeparator()
        self.menu_File.addAction(self.action_Quit)
        self.menu_Edit.addAction(self.action_Copy)
        self.menu_Edit.addAction(self.action_Paste)
        self.menu_Windows.addAction(self.action_Sutta_Search)
        self.menu_Windows.addAction(self.action_Dictionary_Search)
        self.menu_Windows.addAction(self.action_Document_Reader)
        self.menu_Windows.addAction(self.action_Library)
        self.menu_Windows.addAction(self.action_Notes)
        self.menu_Help.addAction(self.action_Website)
        self.menu_Help.addAction(self.action_About)
        self.menu_Library.addAction(self.action_Open_Selected)
        self.menu_Library.addAction(self.action_Add)
        self.menu_Library.addAction(self.action_Remove)
        self.menubar.addAction(self.menu_File.menuAction())
        self.menubar.addAction(self.menu_Edit.menuAction())
        self.menubar.addAction(self.menu_Windows.menuAction())
        self.menubar.addAction(self.menu_Library.menuAction())
        self.menubar.addAction(self.menu_Help.menuAction())
        self.toolBar.addAction(self.action_Sutta_Search)
        self.toolBar.addAction(self.action_Dictionary_Search)
        self.toolBar_2.addAction(self.action_Open_Selected)
        self.toolBar_2.addAction(self.action_Add)
        self.toolBar_2.addAction(self.action_Remove)

        self.retranslateUi(LibraryBrowserWindow)
        QtCore.QMetaObject.connectSlotsByName(LibraryBrowserWindow)

    def retranslateUi(self, LibraryBrowserWindow):
        _translate = QtCore.QCoreApplication.translate
        LibraryBrowserWindow.setWindowTitle(_translate("LibraryBrowserWindow", "Library - Simsapa"))
        self.search_button.setText(_translate("LibraryBrowserWindow", "Search"))
        self.doc_cover.setText(_translate("LibraryBrowserWindow", "Cover"))
        self.doc_title.setText(_translate("LibraryBrowserWindow", "Title"))
        self.doc_author.setText(_translate("LibraryBrowserWindow", "Author"))
        self.menu_File.setTitle(_translate("LibraryBrowserWindow", "&File"))
        self.menu_Edit.setTitle(_translate("LibraryBrowserWindow", "&Edit"))
        self.menu_Windows.setTitle(_translate("LibraryBrowserWindow", "&Windows"))
        self.menu_Help.setTitle(_translate("LibraryBrowserWindow", "&Help"))
        self.menu_Library.setTitle(_translate("LibraryBrowserWindow", "&Library"))
        self.toolBar.setWindowTitle(_translate("LibraryBrowserWindow", "toolBar"))
        self.toolBar_2.setWindowTitle(_translate("LibraryBrowserWindow", "toolBar_2"))
        self.action_Copy.setText(_translate("LibraryBrowserWindow", "&Copy"))
        self.action_Paste.setText(_translate("LibraryBrowserWindow", "&Paste"))
        self.action_Quit.setText(_translate("LibraryBrowserWindow", "&Quit"))
        self.action_Quit.setShortcut(_translate("LibraryBrowserWindow", "Ctrl+Q"))
        self.action_Sutta_Search.setText(_translate("LibraryBrowserWindow", "&Sutta Search"))
        self.action_Sutta_Search.setShortcut(_translate("LibraryBrowserWindow", "F5"))
        self.action_Dictionary_Search.setText(_translate("LibraryBrowserWindow", "&Dictionary Search"))
        self.action_Dictionary_Search.setShortcut(_translate("LibraryBrowserWindow", "F6"))
        self.action_About.setText(_translate("LibraryBrowserWindow", "&About"))
        self.action_Website.setText(_translate("LibraryBrowserWindow", "&Website"))
        self.action_Close_Window.setText(_translate("LibraryBrowserWindow", "&Close Window"))
        self.action_Open.setText(_translate("LibraryBrowserWindow", "&Open..."))
        self.action_Open.setShortcut(_translate("LibraryBrowserWindow", "Ctrl+O"))
        self.action_Document_Reader.setText(_translate("LibraryBrowserWindow", "D&ocument Reader"))
        self.action_Document_Reader.setToolTip(_translate("LibraryBrowserWindow", "Document Reader"))
        self.action_Document_Reader.setShortcut(_translate("LibraryBrowserWindow", "F7"))
        self.action_Add.setText(_translate("LibraryBrowserWindow", "&Add..."))
        self.action_Remove.setText(_translate("LibraryBrowserWindow", "&Remove"))
        self.action_Remove.setShortcut(_translate("LibraryBrowserWindow", "Del"))
        self.action_Open_Selected.setText(_translate("LibraryBrowserWindow", "&Open Selected"))
        self.action_Library.setText(_translate("LibraryBrowserWindow", "&Library"))
        self.action_Library.setShortcut(_translate("LibraryBrowserWindow", "F8"))
        self.action_Notes.setText(_translate("LibraryBrowserWindow", "&Notes"))
        self.action_Notes.setToolTip(_translate("LibraryBrowserWindow", "Notes"))
        self.action_Notes.setShortcut(_translate("LibraryBrowserWindow", "F9"))
from simsapa.assets import icons_rc
